import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:grpc/grpc.dart';
import 'package:source_gen/source_gen.dart';

// Get the service name by removing the suffix 'ServiceBase'
// from the class name of the service.
String getServiceName(ClassElement clazz) {
  final className = clazz.name;
  final classNameLength = className.length;
  final serviceName =
      className.substring(0, classNameLength - "ServiceBase".length);
  return serviceName;
}

// Set the out id.
// Since the 'changeExtension' doesn't know which '.' to split the extension from,
// we are doing it by hand.
// By simply calculate the `inExtension` length and remove it from the path.
AssetId getOutId(AssetId assetId, String inExtension, String outExtension) {
  final path =
      assetId.path.substring(0, assetId.path.length - inExtension.length) +
          outExtension;
  return AssetId(assetId.package, path);
}

// Loop the classes in the library and search for 'Service' generated classes
// to apply the logic into it.
List<ClassElement> getServiceClasses(LibraryReader libraryReader) {
  final classes = <ClassElement>[];
  for (final clazz in libraryReader.classes) {
    if (!TypeChecker.fromRuntime(Service).isSuperOf(clazz)) continue;
    classes.add(clazz);
  }
  return classes;
}

/// Generate abstract implementation for the `methods` provided.
///
/// for example:
/// Future<String> getName();
String methodsToAbstractImplementation(List<MethodElement> methods) {
  final buffer = StringBuffer();
  for (final method in methods) {
    final methodName = method.name;
    if (methodName.endsWith("_Pre")) continue;

    final methodBuffer = StringBuffer("\t");

    // Get the return type.
    // And also replace `Future` to `FutureOr` since it make implementation easier
    // by removing the need to to add `async` when it's not needed.
    final String returnType;
    if (method.returnType.isDartAsyncFuture) {
      returnType =
          method.returnType.toString().replaceFirst("Future", "FutureOr");
    } else {
      returnType = method.returnType.toString();
    }
    methodBuffer.write("$returnType $methodName(");

    // Get the methods.
    // And remove the `ServiceCall` parameter.
    final parameters = <ParameterElement>[];
    for (final parameter in method.parameters) {
      if (TypeChecker.fromRuntime(ServiceCall).isExactlyType(parameter.type))
        continue;
      parameters.add(parameter);
    }

    for (int i = 0; i < parameters.length; i++) {
      final parameter = parameters[i];

      methodBuffer.write("${parameter.type} ${parameter.name}");
      if (i < parameters.length - 1) {
        methodBuffer.write(",");
      }
    }
    methodBuffer.write(");");
    buffer.writeln(methodBuffer);
  }
  return buffer.toString();
}

/// Bind methods to the `binder` by calling the same method name.
///
/// for example:
/// FutureOr<String> getName() async => await _binder.getName();
String bindMethods(List<MethodElement> methods, String binder,
    {List<Type>? excludeParameters, bool returnFutureOr = true}) {
  excludeParameters ??= [];

  final buffer = StringBuffer();
  for (final method in methods) {
    final methodName = method.name;
    if (methodName.endsWith("_Pre")) continue;

    final methodBuffer = StringBuffer();

    methodBuffer.writeln("\t@override");
    methodBuffer.write("\t");

    // Get the return type.
    // And also replace `Future` to `FutureOr` since it make implementation easier
    // by removing the need to to add `async` when it's not needed.
    final String returnType;
    if (returnFutureOr && method.returnType.isDartAsyncFuture) {
      returnType =
          method.returnType.toString().replaceFirst("Future", "FutureOr");
    } else {
      returnType = method.returnType.toString();
    }
    methodBuffer.write("$returnType $methodName(");

    // Get the methods.
    // And remove the `ServiceCall` parameter.
    final parameters = <ParameterElement>[];
    for (final parameter in method.parameters) {
      bool exclude = false;
      for (final excludeParameter in excludeParameters) {
        if (TypeChecker.fromRuntime(excludeParameter)
            .isExactlyType(parameter.type)) {
          exclude = true;
          break;
        }
      }

      if (!exclude) {
        parameters.add(parameter);
      }
    }

    for (int i = 0; i < parameters.length; i++) {
      final parameter = parameters[i];

      methodBuffer.write("${parameter.type} ${parameter.name}");
      if (i < parameters.length - 1) {
        methodBuffer.write(",");
      }
    }
    methodBuffer.write(")");

    if (method.returnType.isDartAsyncFuture) {
      methodBuffer.write("async ");
    }

    methodBuffer.write("=>");

    if (method.returnType.isDartAsyncFuture) {
      methodBuffer.write("await ");
    }

    methodBuffer.write("$binder.${method.name}(request);");

    buffer.writeln(methodBuffer);
  }
  return buffer.toString();
}
